import numpy as np
import rclpy
from rcl_interfaces.msg import ParameterDescriptor, ParameterType
from rclpy.node import Node
from rclpy.qos import qos_profile_sensor_data
from sensor_msgs.msg import Joy as JoyMsg
from std_msgs.msg import Float32, Int32
from std_srvs.srv import Empty

from ls2n_drone_command_center.utilities.utils import load_yaml
from ls2n_interfaces.msg import KeepAlive

qos_profile_sensor_data.depth = 1

# Default indexes for joystick axes and buttons
# (Checked with gamepad EasySMX, might need to be verified if using different brands)
# axes:
vx = 3  # rv (right vertical)
vy = 2  # rh (right horizontal)
vz = 1  # lv
v_yaw = 0  # lh

# update rate (Hz)
RATE = 50


class Joystick(Node):
    joy_config: dict

    def __init__(self):
        super().__init__("Joystick_Node")
        self.get_logger().info("Starting joystick connection")
        joy_config_version_desc = ParameterDescriptor(
            type=ParameterType.PARAMETER_STRING,
            description="Joystick Configuration To Use",
        )
        self.declare_parameter(
            "joy_config_version",
            "CONFIG_1",
            descriptor=joy_config_version_desc,
        )
        joy_config_version = self.get_parameter("joy_config_version").value  # type: ignore

        self.joy_config = load_yaml(filename="joystick_config")[
            joy_config_version
        ]
        # Subscription to joytick inputs
        self.joy_sub = self.create_subscription(
            JoyMsg,
            "/CommandCenter/Joystick",
            self.joystick_callback,
            qos_profile_sensor_data,
        )
        # Keep alive
        self.keep_alive_publisher = self.create_publisher(
            KeepAlive, "KeepAlive", qos_profile_sensor_data
        )
        self.create_timer(1.0 / RATE, self.keep_alive_callback)
        self.keep_alive = False
        # Service clients
        self.spin_motor_client = self.create_client(Empty, "SpinMotors")
        self.button_A = 0
        self.start_experiment_client = self.create_client(
            Empty, "StartExperiment"
        )
        self.button_B = 0
        self.stop_experiment_client = self.create_client(
            Empty, "StopExperiment"
        )
        self.button_X = 0
        # Broadcast of the axis values
        self.lh_pub = self.create_publisher(
            Float32, "Joystick/LeftHorizontal", qos_profile_sensor_data
        )
        self.lv_pub = self.create_publisher(
            Float32, "Joystick/LeftVertical", qos_profile_sensor_data
        )
        self.rh_pub = self.create_publisher(
            Float32, "Joystick/RightHorizontal", qos_profile_sensor_data
        )
        self.rv_pub = self.create_publisher(
            Float32, "Joystick/RightVertical", qos_profile_sensor_data
        )

        self.btn_a_pub = self.create_publisher(
            Int32, "Joystick/Button/A", qos_profile_sensor_data
        )
        self.btn_b_pub = self.create_publisher(
            Int32, "Joystick/Button/B", qos_profile_sensor_data
        )
        self.btn_x_pub = self.create_publisher(
            Int32, "Joystick/Button/X", qos_profile_sensor_data
        )
        self.btn_y_pub = self.create_publisher(
            Int32, "Joystick/Button/Y", qos_profile_sensor_data
        )
        self.btn_rb_pub = self.create_publisher(
            Int32, "Joystick/Button/RB", qos_profile_sensor_data
        )
        self.btn_lb_pub = self.create_publisher(
            Int32, "Joystick/Button/LB", qos_profile_sensor_data
        )
        self.btn_rt_pub = self.create_publisher(
            Float32, "Joystick/Button/RT", qos_profile_sensor_data
        )
        self.btn_lt_pub = self.create_publisher(
            Float32, "Joystick/Button/LT", qos_profile_sensor_data
        )
        self.btn_back_pub = self.create_publisher(
            Int32, "Joystick/Button/Back", qos_profile_sensor_data
        )
        self.btn_start_pub = self.create_publisher(
            Int32, "Joystick/Button/Start", qos_profile_sensor_data
        )

        self.hp_pub = self.create_publisher(
            Int32, "Joystick/Axes/HorizontalPad", qos_profile_sensor_data
        )
        self.lp_pub = self.create_publisher(
            Int32, "Joystick/Axes/VerticalPad", qos_profile_sensor_data
        )

    def right_trigger(self, *, data):
        rt = self.joy_config["RT"]
        if rt["TYPE"] == "button":
            return float(self.get_joy_data(data=data, input_pt="RT"))
        elif rt["TYPE"] == "axis":
            return -np.round(self.get_joy_data(data=data, input_pt="RT"))

    def left_trigger(self, *, data):
        lt = self.joy_config["LT"]
        if lt["TYPE"] == "button":
            return float(
                self.get_joy_data(data=data, input_pt="LT")
            )  # left trigger
        elif lt["TYPE"] == "axis":
            return np.round(
                (1 - self.get_joy_data(data=data, input_pt="LT") / 2, 3)
            )  # left trigger

    def get_joy_data(self, *, data, input_pt: str):
        input_data = self.joy_config[input_pt]
        if input_data["TYPE"] == "axis":
            return data.axes[input_data["INDEX"]]
        elif input_data["TYPE"] == "button":
            return data.buttons[input_data["INDEX"]]

    def joystick_callback(self, data):
        lh = np.round(
            1 * self.get_joy_data(data=data, input_pt="LH"), 3
        )  # left horizontal
        lv = np.round(
            self.get_joy_data(data=data, input_pt="LV"), 3
        )  # left vertical
        rh = np.round(
            1 * self.get_joy_data(data=data, input_pt="RH"), 3
        )  # right horizontal
        rv = np.round(
            self.get_joy_data(data=data, input_pt="RV"), 3
        )  # right vertical
        lt = self.left_trigger(data=data)  # left trigger
        rt = self.right_trigger(data=data)
        hp = -int(
            self.get_joy_data(data=data, input_pt="HP")
        )  # horizontal pad
        vp = int(self.get_joy_data(data=data, input_pt="VP"))  # vertical pad
        Ab = self.get_joy_data(data=data, input_pt="A")  # A Button
        Bb = self.get_joy_data(data=data, input_pt="B")  # B Button
        Xb = self.get_joy_data(data=data, input_pt="X")  # X Button
        Yb = self.get_joy_data(data=data, input_pt="Y")  # Y Button
        lb = self.get_joy_data(data=data, input_pt="LB")  # left trigger Button
        rb = self.get_joy_data(
            data=data, input_pt="RB"
        )  # right trigger Button
        bk = self.get_joy_data(data=data, input_pt="BACK")  # Back Button
        st = self.get_joy_data(data=data, input_pt="START")  # Start Button

        if rt == 1.0:  # KeepAlive
            self.keep_alive = True
        else:
            self.keep_alive = False
        # Buttons are activated on realease
        # button A for spinning motors
        if Ab == 1 and self.button_A == 0:
            self.get_logger().info("Spin Motors")
            self.spin_motor()
        # button B for starting experiment
        if Bb == 1 and self.button_B == 0:
            self.get_logger().info("Start Experiment")
            self.start_experiment()
        # button B for stopping experiment
        if Xb == 1 and self.button_X == 0:
            self.get_logger().info("Stop Experiment")
            self.stop_experiment()

        self.button_A = Ab
        self.button_B = Bb
        self.button_X = Xb

        # Published the normalized axis data
        self.lh_pub.publish(Float32(data=lh))
        self.rh_pub.publish(Float32(data=rh))
        self.lv_pub.publish(Float32(data=lv))
        self.rv_pub.publish(Float32(data=rv))

        self.btn_a_pub.publish(Int32(data=Ab))
        self.btn_b_pub.publish(Int32(data=Bb))
        self.btn_x_pub.publish(Int32(data=Xb))
        self.btn_y_pub.publish(Int32(data=Yb))

        self.btn_rb_pub.publish(Int32(data=rb))
        self.btn_lb_pub.publish(Int32(data=lb))

        self.btn_rt_pub.publish(Float32(data=rt))
        self.btn_lt_pub.publish(Float32(data=lt))

        self.btn_back_pub.publish(Int32(data=bk))
        self.btn_start_pub.publish(Int32(data=st))

        self.hp_pub.publish(Int32(data=hp))
        self.lp_pub.publish(Int32(data=vp))

    def keep_alive_callback(self):
        if self.keep_alive:
            msg = KeepAlive()
            msg.keep_alive = True
            msg.feedback = False
            msg.stamp = self.get_clock().now().to_msg()
            self.keep_alive_publisher.publish(msg)

    def spin_motor(self):
        self.spin_motor_client.call_async(Empty.Request())

    def start_experiment(self):
        self.start_experiment_client.call_async(Empty.Request())

    def stop_experiment(self):
        self.stop_experiment_client.call_async(Empty.Request())


def main(args=None):
    rclpy.init(args=args)
    joystick = Joystick()
    try:
        rclpy.spin(joystick)
    except KeyboardInterrupt:
        print("Shutting down joystick connection")
    joystick.destroy_node()
    rclpy.try_shutdown()


if __name__ == "main":
    main()
